/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 * Author: A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2007 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION --
// This class is the abstract class that serves as the base of all the
// Pdf builders of the RhoPiTools/RhoRhoTools packages. To write a Pdf 
// builder, one should :
//
// - inherit from AFitAbsPdfBuilder
//
// - define a constructor that is meant to be initialised with the 
// user's discriminating variables. At the end of the constructor, 
// readDataCard() should be called.
//
// - implement the buildPdf() method that will build the Pdf pointed to by
// _thePdf. During the building, it is possible to read the flag saying 
// wether the variables should be blind by testing isData().
//
// - implement readDataCard() and writeDataCard(ostream& out) that will 
// read and write the parameters contained by the builder in a section
// specific to it.

#include <assert.h>
#include <string.h>

#include "AFit/AFitAbsPdfBuilder.hh"
#include "TMath.h"

ClassImp(AFitAbsPdfBuilder)
  ;

RooArgSet AFitAbsPdfBuilder::globalConfig;

RooRealVar AFitAbsPdfBuilder::zero("zero", "0", 0.);
RooRealVar AFitAbsPdfBuilder::plusone("plusone", "+1", 1.);
RooRealVar AFitAbsPdfBuilder::minusone("minusone", "-1", -1.);
RooRealVar AFitAbsPdfBuilder::onehalf("onehalf", "0.5", 0.5);
RooRealVar AFitAbsPdfBuilder::pi("pi", "pi", TMath::Pi() );

Bool_t AFitAbsPdfBuilder::isDataFlag = false;

AFitAbsPdfBuilder::AFitAbsPdfBuilder() :
 writtenToConfigFile(kFALSE)
{
  // Constructor
  _thePdf = 0;
  _discVars = 0;
  _deleteThePdf = true;
  //  setVerbose(kFALSE);
}

AFitAbsPdfBuilder::~AFitAbsPdfBuilder() {
  // Destructor
  if (_deleteThePdf && _thePdf) delete _thePdf;
}

RooAbsPdf* AFitAbsPdfBuilder::getPdf() {
  // Return the built Pdf
  //  if(_verbose) cout << ClassName() << ": called getPdf()" << endl;
  if (!_thePdf) {
    cout << ClassName() << _getPdfStr << endl;
    buildPdf();
    if(!_thePdf) {
      cout << ClassName() << ": FATAL ERROR unable to build PDF" << endl; 
      assert(_thePdf); 
    }
    cout << " for the PDF called " << _thePdf->GetName() << endl;  
  }
  //  if(_verbose) cout << ClassName() << ": returning fromgetPdf()" << endl;
  return _thePdf;
}

char* AFitAbsPdfBuilder::dataCard = 0;

void AFitAbsPdfBuilder::setDataCard(const char* init) {
  // Static method to set the name of the file to be read for the data card.
  // Before the user does anything, he should make a call such as :
  // AFitAbsPdfBuilder::setDataCard("myDataCard.txt");
  if (init) {
    dataCard = new char[strlen(init)+1];
    strcpy(dataCard, init);
  } else {
    dataCard = 0;
  }
}

const char* AFitAbsPdfBuilder::getDataCard() {
  // Return the name of the file containing the data card.
  return dataCard;
}

RooRealVar* AFitAbsPdfBuilder::makeRealVar(const RooRealVar& var) {
  RooRealVar* res;
  RooAbsArg * arg = globalConfig.find(var.GetName());
  if (!arg) {
    res = new RooRealVar(var);
    globalConfig.add(*res);
    return res;
  } else {
    if (res = dynamic_cast<RooRealVar*>(arg)) return res;
    else {
      cout << "AFitAbsPdfBuilder::makeRealVar: Found argument of name "
	   << var.GetName()
	   << " which is not a RooRealVar*" << endl;
      arg->Print("v");
      abort();
    }
  }
}

void AFitAbsPdfBuilder::fixAllParams(Bool_t fixParams) {
  /*
   * Function to switch between fixing and floating all of the
   * parameters in a PDF associated with this model.
   */
  RooAbsPdf * pdf = getPdf();
  RooArgSet * compos = pdf->getComponents();
  RooArgSet * pdfpar = pdf->getParameters(compos);
 
  TIterator * iter = pdfpar->createIterator();
  TIter next(iter);
  TObject * obj=0;
  while (obj = (TObject *)next())  {
    RooRealVar * var = dynamic_cast<RooRealVar*>(obj);
    if(var) var->setConstant(fixParams);
  }
}

void AFitAbsPdfBuilder::readDataCard() {
  const char * datacard = getDataCard();
  if(datacard) {
    ifstream in(getDataCard());
    if(!in) {
      cout << ClassName() << ": ERROR - unable to open datacard " << datacard << endl;
    } else cout << ClassName() << ": reading data card " << datacard << endl; 
  } else {
    // data card has not been set, so do nothing
    return;
  }
  varSet.readFromFile(getDataCard(), 0, _name);
}

void AFitAbsPdfBuilder::writeDataCard(ostream& out) {
  if(!writtenToConfigFile) {
    out << "["<< _name << "]" << endl;
    varSet.writeToStream(out, false);
    writtenToConfigFile = kTRUE;
  }
}
