 /*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 *    File: $Id: AFitTMVAInterface.cc,v 1.5 2005/07/03 23:04:22 bevan Exp $
 * Authors:
 *    A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2009 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [AFitTMVAInterface] --
// This class is an interface for using TMVA to train a MVA for use
// in analysis.  Also can be used to make plots etc summarising the 
// differences in MVA performance if more than one method is chosen
//
// Please see the TMVA documentation for more details.  This
// can be obtained at the following URL:
//    
//                http://tmva.sourceforge.net/
//
#include "AFit/AFitTMVAInterface.hh"

#ifdef __BABARROOFIT__
#include "RooFitCore/RooDataSet.hh"
#else
#include "RooDataSet.h"
#endif

#include "TFile.h"
#include "TObjArray.h"
#include "TSystem.h"
#include "TROOT.h"

#include "TMVA/Factory.h"

#include <fstream>
#include <map>

using namespace std;

ClassImp(AFitTMVAInterface)
  ;

class reader_variable{
public:
  reader_variable() { dbl=0.0; flt=0.0; };
  ~reader_variable() {};

  Double_t dbl; // from TTree made using a RooDataSet
  Float_t flt;  // required by TMVA
};

AFitTMVAInterface::AFitTMVAInterface(TString configFile) 
  :   sigFile("sigFile", "Signal ROOT File", "tmva_signal.root")
  , bgFile("bgFile", "Background ROOT File", "tmva_background.root")
  , dataName("dataName", "Data KEY name", "data")
  , outputFileName("outputFileName", "Output File Name", "tmva_out.root")
  , trainingMethod("trainingMethod", "List of training methods", "Fisher")
  , variables("variables", "List of variables for MVA", "null")
  , factoryOptions("factoryOptions", "factoryOptions", "")
  , trainingOptions("trainingOptions", "trainingOptions", "")
  , preselectionCut("preselectionCut", "preselectionCut", "")
  , weightsdir("weightsdir", "Weights directory", "weights")
  , analysisprefix("analysisprefix", "Analysis prefix name", "AFitTMVAInterface-TMVAanalysis")
  , Cuts_method("Cuts", "TMVA Cuts Method", 0.0)
  , CutsD_method("CutsD", "TMVA CutsD Method", 0.0)
  , CutsGA_method("CutsGA", "TMVA CutsGA Method", 0.0)
  , Likelihood_method("Likelihood", "TMVA Likelihood Method", 0.0)
  , LikelihoodD_method("LikelihoodD", "TMVA LikelihoodD Method", 0.0)
  , LikelihoodPCA_method("LikelihoodPCA", "TMVA LikelihoodPCA Method", 0.0)
  , PDERS_method("PDERS", "TMVA PDERS Method", 0.0)
  , PDERSD_method("PDERSD", "TMVA PDERSD Method", 0.0)
  , PDERSPCA_method("PDERSPCA", "TMVA PDERSPCA Method", 0.0)
  , KNN_method("KNN", "TMVA KNN Method", 0.0)
  , HMatrix_method("HMatrix", "TMVA HMatrix Method", 0.0)
  , Fisher_method("Fisher", "TMVA Fisher Method", 0.0)
  , MLP_method("MLP", "TMVA MLP Method", 0.0)
  , CFMlpANN_method("CFMlpANN", "TMVA CFMlpANN Method", 0.0)
  , TMlpANN_method("TMlpANN", "TMVA TMlpANN Method", 0.0)
  , BDT_method("BDT", "TMVA BDT Method", 0.0)
  , BDTD_method("BDTD", "TMVA BDTD Method", 0.0)
  , RuleFit_method("RuleFit", "TMVA RuleFit Method", 0.0)
  , SVM_method("SVM", "TMVA SVM Method", 0.0)
  , SVM_Gauss_method("SVM_Gauss", "TMVA SVM_Gauss Method", 0.0)
  , SVM_Poly_method("SVM_Poly", "TMVA SVM_Poly Method", 0.0)
  , SVM_Lin_method("SVM_Lin", "TMVA SVM_Lin Method", 0.0)
  , FDA_MT_method("FDA_MT", "TMVA FDA_MT Method", 0.0)
  , FDA_GA_method("FDA_GA", "TMVA FDA_GA Method", 0.0)
{
  // Constructor

  _cfgFile = configFile;
  /*
   * Read in options from the configuration file, and then run the MVA 
   * analysis.
   */

  varSet.add(sigFile);
  varSet.add(bgFile);
  varSet.add(dataName);
  varSet.add(outputFileName);
  varSet.add(trainingMethod);
  varSet.add(variables);
  varSet.add(factoryOptions);
  varSet.add(trainingOptions);

  ifstream in(_cfgFile);
  if(!in) {
    cout << ClassName() << ": ERROR - unable to open datacard " << _cfgFile << endl;
  } else cout << ClassName() << ": reading data card " << _cfgFile << endl; 

  varSet.readFromFile(_cfgFile, 0, "TMVAInterface");
}

void AFitTMVAInterface::trainMethods()
{
  /*
   * Set up the signal and background reference files
   */
  RooRealVar sigWeight("sigWeight", "Signal Tree Weight", 1.0);
  RooRealVar bgWeight("bgWeight", "Background Tree Weight", 1.0);

  cout << "Using the signal file:     " << sigFile.getVal() << endl;
  cout << "Using the background file: " << bgFile.getVal() << endl;
  TFile fsig(sigFile.getVal());
  TFile fbg(bgFile.getVal());

  RooDataSet * dsig = (RooDataSet*)fsig.Get(dataName.getVal());
  RooDataSet * dbg  = (RooDataSet*)fbg.Get(dataName.getVal());
  TTree * tsig = (TTree*)(dsig->tree().Clone("data")); 
  TTree * tbg = (TTree*)(dbg->tree().Clone("data")); 

  if(!tsig || !tbg) {
    cout << "Unable to get tree with name " << dataName.getVal() << " from file(s):" << endl;
    if(!tsig ) cout<< "\t" << sigFile.getVal() << endl;
    if(!tbg )  cout<< "\t" << bgFile.getVal() << endl;
    return;
  }

  cout << "\tSignal tree has " << tsig->GetEntries() << " entries" << endl;
  cout << "\tBackground tree has " << tbg->GetEntries() << " entries" << endl;

  /*
   * Set up the TMVA factory to run the training
   */
  TFile * fout = new TFile(outputFileName.getVal(), "RECREATE");
  //  TMVA::Factory *factory = new TMVA::Factory(analysisprefix.getVal(), fout, factoryOptions.getVal()); //Form("!V:!Silent:%sColor", gROOT->IsBatch()?"!":"") );
  TMVA::Factory factory(analysisprefix.getVal(), fout, factoryOptions.getVal()); //Form("!V:!Silent:%sColor", gROOT->IsBatch()?"!":"") );

  factory.AddSignalTree( tsig, sigWeight.getVal() );
  factory.AddBackgroundTree( tbg, bgWeight.getVal() );

  /*
   * Configure Training Options
   */

  cout.width(80);
  cout.fill('_');
  cout << ""<< endl;
  cout << "\tBooking variables for the TMVA::Factory" << endl;
  cout.width(80);
  cout.fill('_');
  cout << ""<< endl<< endl;
  // set the list of variables to use
  TObjArray * varArray = TString(variables).Tokenize(",");
  for(Int_t j=0; j < varArray->GetEntries(); j++) {
    TString thisVar = ((TObjString*)varArray->At(j))->GetString();
    TObjArray * thisVarArray = thisVar.Tokenize(":");
    if(thisVarArray->GetEntries() == 2) {
      TString jVariable = ((TObjString*)thisVarArray->At(0))->GetString();
      TString jType      = (char*)(((TObjString*)thisVarArray->At(1))->GetString()).Data();
      factory.AddVariable(jVariable, (char)jType.Data()[0]);
    } else {
      cout << ClassName()  << " unable to append variable to factory list: " << thisVar << endl;
    }
  }

  TCut preselectionCut = "";
  factory.PrepareTrainingAndTestTree(preselectionCut, trainingOptions.getVal());

  /*
   * The following list of types is taken from TMVA/Types.h
   *
   *     kVariable
   *     kCuts
   *     kLikelihood
   *     kPDERS
   *     kHMatrix
   *     kFisher
   *     kCFMlpANN
   *     kTMlpANN
   *     kBDT
   *     kRuleFit
   *     kSVM
   *     kMLP
   *     kBayesClassifier,
   */
  cout.width(80);
  cout.fill('_');
  cout << ""<< endl;
  cout << "\tBooking the methods for the TMVA::Factory" << endl;
  cout.width(80);
  cout.fill('_');
  cout << ""<< endl<< endl;
  cout << "Booking the training method(s)" << endl;
  TString _trainingMethod = trainingMethod; 
  if(_trainingMethod.Contains("Cuts"))            factory.BookMethod(TMVA::Types::kCuts, "Cuts");
  if(_trainingMethod.Contains("Likelihood"))      factory.BookMethod(TMVA::Types::kLikelihood, "Likelihood");
  //  if(_trainingMethod.Contains("PDERS"))           factory.BookMethod(TMVA::Types::kPDERS, "PDERS");
  if(_trainingMethod.Contains("HMatrix"))         factory.BookMethod(TMVA::Types::kHMatrix, "HMatrix");
  if(_trainingMethod.Contains("Fisher"))          factory.BookMethod(TMVA::Types::kFisher, "Fisher");
  if(_trainingMethod.Contains("MLP"))             factory.BookMethod(TMVA::Types::kMLP, "MLP");
  if(_trainingMethod.Contains("CFMlpANN"))        factory.BookMethod(TMVA::Types::kCFMlpANN, "CFMlpANN");
  if(_trainingMethod.Contains("TMlpANN"))         factory.BookMethod(TMVA::Types::kTMlpANN, "TMlpANN");
  if(_trainingMethod.Contains("BDT"))             factory.BookMethod(TMVA::Types::kBDT, "BDT");
  if(_trainingMethod.Contains("RuleFit"))         factory.BookMethod(TMVA::Types::kRuleFit, "RuleFit");
  if(_trainingMethod.Contains("SVM"))             factory.BookMethod(TMVA::Types::kSVM, "SVM");
  if(_trainingMethod.Contains("BayesClassifier")) factory.BookMethod(TMVA::Types::kBayesClassifier, "BayesClassifier");
  //if(_trainingMethod.Contains("Committee"))       factory.BookMethod(TMVA::Types::kCommittee, "Committee");
  //if(_trainingMethod.Contains("MaxMethod"))       factory.BookMethod(TMVA::Types::kMaxMethod, "MaxMethod");

  cout << "Training the booked method(s)" << endl;
  cout << "_______________________________________________________" << endl;
  factory.TrainAllMethods();

  /*
   * Train MVAs, prepare and output results
   */
  cout << "Testing all methods" << endl;
  cout << "_______________________________________________________" << endl;
  factory.TestAllMethods();

  //cout << "Evaluating all variables" << endl;
  //cout << "_______________________________________________________" << endl;
  //  factory.EvaluateAllVariables();

  cout << "Evaluating all methods" << endl;
  cout << "_______________________________________________________" << endl;
  factory.EvaluateAllMethods();

  cout << "_______________________________________________________" << endl;
  cout << "\tMVA training has finished." << endl; 
  cout << "_______________________________________________________" << endl;

  //  delete factory;
}

void AFitTMVAInterface::runReader(TString sourcefile, TString targetfile) {
  /*
   * Once the MVAs have been computed, this function facilitates adding those 
   * MVA values to data sets for one or more MVA methods.  The original data set
   * remains unaltered, and a new one is created with the additional variables.
   *
   * The MVA variable names are the same as the keys used by this class: 
   *   e.g Fisher, MLP, etc.
   */
  TFile fin(sourcefile);
  RooDataSet * data = (RooDataSet*)fin.Get("data");

  RooDataSet * newDataSet = runReader(data);

  cout.width(80);
  cout.fill('_');
  cout << " "<< endl;
  cout << "\tWriting data set to file (with all booked MVAs added to the set" << endl;
  cout.width(80);
  cout.fill('_');
  cout << ""<< endl<< endl;
  cout.fill(' ');

  newDataSet->Print("v");

  TFile fout(targetfile, "RECREATE");
  newDataSet->Write();
  fout.Write();
  fout.Close();
}

RooDataSet * AFitTMVAInterface::runReader(RooDataSet * data) {
  /*
   * Once the MVAs have been computed, this function facilitates adding those 
   * MVA values to data sets for one or more MVA methods.  The original data set
   * remains unaltered, and a new one is created with the additional variables.
   *
   * The MVA variable names are the same as the keys used by this class: 
   *   e.g Fisher, MLP, etc.
   */
  TMVA::Reader * reader = new TMVA::Reader();
  TTree * tree = (TTree*)(data->tree().Clone("data"));

  // set up a map of reader_variables for use by TMVA and the RooDataSet's tree
  map<TString, reader_variable*> theMap;

  cout.width(80);
  cout.fill('_');
  cout << ""<< endl;
  cout << "\tBooking variables with the TMVA::Reader" << endl;
  cout.width(80);
  cout.fill('_');
  cout << ""<< endl<< endl;
  // make variables variables to use in the readback
  TObjArray * varArray = TString(variables).Tokenize(",");
  for(Int_t j=0; j < varArray->GetEntries(); j++) {
    TString thisVar = ((TObjString*)varArray->At(j))->GetString();
    TObjArray * thisVarArray = thisVar.Tokenize(":");
    if(thisVarArray->GetEntries() == 2) {
      TString jVariable = ((TObjString*)thisVarArray->At(0))->GetString();
      TString jType      = (char*)(((TObjString*)thisVarArray->At(1))->GetString()).Data();

      reader_variable * thisRV = new reader_variable;
      theMap.insert(make_pair(jVariable, thisRV));

      reader->AddVariable(jVariable, &thisRV->flt);
      tree->SetBranchAddress(jVariable, &thisRV->dbl);

      if(jType=="I") {
	Int_t * thisVar = new Int_t;
	reader->AddVariable(jVariable, thisVar);
	tree->SetBranchAddress(jVariable, thisVar);
       }
    } else {
      cout << ClassName()  << " unable to append variable to factory list: " << thisVar << endl;
    }
  }

  TString methodLocPrefix = weightsdir.getVal();
  methodLocPrefix += "/";
  methodLocPrefix += analysisprefix.getVal();
  cout << "Location of weights files = " << methodLocPrefix << "*" << endl;
  cout << "The following MVA method(s) have been booked: "<< trainingMethod.getVal() << endl;

  cout.width(80);
  cout.fill('_');
  cout << ""<< endl;
  cout << "\tBooking the methods for the TMVA::Reader" << endl;
  cout.width(80);
  cout.fill('_');
  cout << ""<< endl<< endl;

  // book methods for the reader, and making sure that these will be 
  // registered with the output data set
  RooArgSet * newset = (RooArgSet*)data->get(0);

  TString _trainingMethod = trainingMethod; 	  
  if(_trainingMethod.Contains("Cuts"))             { 
    bookReaderMethod(Cuts_method, newset, reader, "Cuts method", methodLocPrefix + "_Cuts.weights.txt" );
    bookReaderMethod(CutsD_method, newset, reader, "CutsD method", methodLocPrefix + "_CutsD.weights.txt" );
    bookReaderMethod(CutsGA_method, newset, reader, "CutsGA method", methodLocPrefix + "_CutsGA.weights.txt" );
  }
  if(_trainingMethod.Contains("Likelihood"))       { 
    bookReaderMethod(Likelihood_method, newset, reader, "Likelihood method", methodLocPrefix + "_Likelihood.weights.txt" );
    bookReaderMethod(LikelihoodD_method, newset, reader, "LikelihoodD method", methodLocPrefix + "_LikelihoodD.weights.txt" );
    bookReaderMethod(LikelihoodPCA_method, newset, reader, "LikelihoodPCA method", methodLocPrefix + "_LikelihoodPCA.weights.txt" );
  }
  if(_trainingMethod.Contains("BDT")) {
    bookReaderMethod(BDT_method, newset, reader, "BDT method", methodLocPrefix + "_BDT.weights.txt" );
    bookReaderMethod(BDTD_method, newset, reader, "BDTD method", methodLocPrefix + "_BDTD.weights.txt" );
  }
  if(_trainingMethod.Contains("SVM")) { 
    bookReaderMethod(SVM_method, newset, reader, "SVM method", methodLocPrefix + "_SVM.weights.txt" );
    bookReaderMethod(SVM_Gauss_method, newset, reader, "SVM_Gauss method", methodLocPrefix + "_SVM_Gauss.weights.txt" );
    bookReaderMethod(SVM_Poly_method, newset, reader, "SVM_Poly method", methodLocPrefix + "_SVM_Poly.weights.txt" );
    bookReaderMethod(SVM_Lin_method, newset, reader, "SVM_Lin method", methodLocPrefix + "_SVM_Lin.weights.txt" );
  }
  if(_trainingMethod.Contains("FDA")) { 
    bookReaderMethod(FDA_MT_method, newset, reader, "FDA_MT method", methodLocPrefix + "_FDA_MT.weights.txt" );
    bookReaderMethod(FDA_GA_method, newset, reader, "FDA_GA method", methodLocPrefix + "_FDA_GA.weights.txt" );
  }
  /*
  if(_trainingMethod.Contains("PDERS")) {
    bookReaderMethod(PDERS_method, newset, reader, "PDERS method", methodLocPrefix + "PDERS_.weights.txt" );
    bookReaderMethod(PDERSD_method, newset, reader, "PDERSD method", methodLocPrefix + "PDERSD_.weights.txt" );
    bookReaderMethod(PDERSPCA_method, newset, reader, "PDERSPCA method", methodLocPrefix + "PDERSPCA_.weights.txt" );
  }
  */
  if(_trainingMethod.Contains("KNN"))        bookReaderMethod(KNN_method, newset, reader, "KNN method", methodLocPrefix + "_KNN.weights.txt" );
  if(_trainingMethod.Contains("HMatrix"))    bookReaderMethod(HMatrix_method, newset, reader, "HMatrix method", methodLocPrefix + "_HMatrix.weights.txt" );
  if(_trainingMethod.Contains("Fisher"))     bookReaderMethod(Fisher_method, newset, reader, "Fisher method", methodLocPrefix + "_Fisher.weights.txt" );
  if(_trainingMethod.Contains("MLP"))        bookReaderMethod(MLP_method, newset, reader, "MLP method", methodLocPrefix + "_MLP.weights.txt" );
  if(_trainingMethod.Contains("CFMlpANN"))   bookReaderMethod(CFMlpANN_method, newset, reader, "CFMlpANN method", methodLocPrefix + "_CFMlpANN.weights.txt" );
  if(_trainingMethod.Contains("TMlpANN"))    bookReaderMethod(TMlpANN_method, newset, reader, "TMlpANN method", methodLocPrefix + "_TMlpANN.weights.txt" );
  if(_trainingMethod.Contains("RuleFit"))    bookReaderMethod(RuleFit_method, newset, reader, "RuleFit method", methodLocPrefix + "_RuleFit.weights.txt" );

  RooDataSet * newDataSet = new RooDataSet("data", "Data with MVA(s) added to it", *newset);

  for(int i=0; i < data->numEntries(); i++) {
    tree->GetEntry(i);
    RooArgSet * aset = (RooArgSet*)data->get(i);

    // make sure that the tree values for MVA variables are copied to the 
    // corresponding TMVA::Reader values.
    map<TString, reader_variable*>::iterator iMap  = theMap.begin();
    for(int i = 0; i < theMap.size(); i++) {
      reader_variable * thisVar = ((*iMap).second);
      TString thisJVar = (*iMap).first;
      RooRealVar * thisJVarRRV = (RooRealVar*)aset->find(thisJVar);
      thisVar->flt = thisJVarRRV->getVal();
      thisVar->dbl = thisJVarRRV->getVal();
      // cout << (*iMap).first << " "  << thisVar->dbl << " " << thisVar->flt << endl;
      iMap++;
    }

    // evaluate all of the methods required, appending the values of the 
    // MVA to the new data set that will be written to the TFile fout at
    // the end of this function.
    if(_trainingMethod.Contains("Cuts"))          fillRRV(Cuts_method, reader, "Cuts method", aset );
    if(_trainingMethod.Contains("CutsD"))         fillRRV(CutsD_method, reader, "CutsD method", aset );
    if(_trainingMethod.Contains("CutsGA"))        fillRRV(CutsGA_method, reader, "CutsGA method", aset );
    if(_trainingMethod.Contains("Likelihood"))    fillRRV(Likelihood_method, reader, "Likelihood method", aset );
    if(_trainingMethod.Contains("LikelihoodD"))   fillRRV(LikelihoodD_method, reader, "LikelihoodD method", aset );
    if(_trainingMethod.Contains("LikelihoodPCA")) fillRRV(LikelihoodPCA_method, reader, "LikelihoodPCA method", aset );
    //if(_trainingMethod.Contains("PDERS"))         fillRRV(PDERS_method, reader, "PDERS method", aset );
    //if(_trainingMethod.Contains("PDERSD"))        fillRRV(PDERSD_method, reader, "PDERSD method", aset );
    //if(_trainingMethod.Contains("PDERSPCA"))      fillRRV(PDERSPCA_method, reader, "PDERSPCA method", aset );
    if(_trainingMethod.Contains("KNN"))           fillRRV(KNN_method, reader, "KNN method", aset );
    if(_trainingMethod.Contains("HMatrix"))       fillRRV(HMatrix_method, reader, "HMatrix method", aset );
    if(_trainingMethod.Contains("Fisher"))        fillRRV(Fisher_method, reader, "Fisher method", aset );
    if(_trainingMethod.Contains("MLP"))           fillRRV(MLP_method, reader, "MLP method", aset );
    if(_trainingMethod.Contains("CFMlpANN"))      fillRRV(CFMlpANN_method, reader, "CFMlpANN method", aset );
    if(_trainingMethod.Contains("TMlpANN"))       fillRRV(TMlpANN_method, reader, "TMlpANN method", aset );
    if(_trainingMethod.Contains("BDT"))           fillRRV(BDT_method, reader, "BDT method", aset );
    if(_trainingMethod.Contains("BDTD"))          fillRRV(BDTD_method, reader, "BDTD method", aset );
    if(_trainingMethod.Contains("RuleFit"))       fillRRV(RuleFit_method, reader, "RuleFit method", aset );
    if(_trainingMethod.Contains("SVM_Gauss"))     fillRRV(SVM_Gauss_method, reader, "SVM_Gauss method", aset );
    if(_trainingMethod.Contains("SVM_Poly"))      fillRRV(SVM_Poly_method, reader, "SVM_Poly method", aset );
    if(_trainingMethod.Contains("SVM_Lin"))       fillRRV(SVM_Lin_method, reader, "SVM_Lin method", aset );
    if(_trainingMethod.Contains("FDA_MT"))        fillRRV(FDA_MT_method, reader, "FDA_MT method", aset );
    if(_trainingMethod.Contains("FDA_GA"))        fillRRV(FDA_GA_method, reader, "FDA_GA method", aset );

    newDataSet->add(*aset);
  }

  return newDataSet;
}

void AFitTMVAInterface::fillRRV(RooRealVar & var, TMVA::Reader * reader, TString method, RooArgSet * aset) {
  /*
   * Evaluate the MVA method, and ensure that the value of this MVA is well behaved.
   * The resulting variable will be added to the data set, through appending the 
   * reader variable 'var' to 'aset'.
   */
  Float_t mva_value = reader->EvaluateMVA( method );
  if(isnan(mva_value)) mva_value = -999.0;
  var.setVal(mva_value);
  aset->add(var);
}

void AFitTMVAInterface::bookReaderMethod(RooRealVar & var, RooArgSet * aset, TMVA::Reader * reader, TString method, TString weightfile) {
  /*
   * Book the TMVA method, making sure that the weight file exists before doing so 
   */
  ifstream in(weightfile);
  if(!in) {
    cout << "Unable to locate weightfile " << weightfile << " - skipping this classifier" << endl;
    return;
  }
  in.close();

  reader->BookMVA( method, weightfile );

  aset->add(var);
  
}

